unit Drives;
{
	Drives Unit for:
		Getting and setting drive labels.
		Getting and setting drive serial number.
		Testing if a drive is ready.
		Determining the type of drive (hard/3.5/5.25...)
		Return last DOS error status.

  All procedures and functions are protected from DOS critical errors.

	Author: William R. Florac
  Company: FITCO, Verona, WI (wee little company from my house)
	Copyright 1996, FITCO.  All rights reserved.

  1) Users of Drives.pas must accept this disclaimer of warranty:
       This Unit is supplied as is.  The Fitco disclaims all
       warranties, expressed or implied, including, without limitation,
       the warranties of merchantability and of fitness for any purpose.
       Fitco assumes no liability for damages, direct or conse-
       quential, which may result from the use of this Unit."

  2) This Unit is donated to the public as public domain except as
     noted below.

  3) You must copy all Software without modification and must include
     all pages, if the Software is distributed without inclusion in your
     software product. If you are incorporating the Software in
     conjunction with and as a part of your software product which adds
     substantial value, you may modify and include portions of the
     Software.

  4) Fitco retains the copyright for this Unit.  You may not distribute
     the source code (PAS) or its compiled unit (DCU) for profit.

  5) If you do find this Unit handy and you feel guilty
     for using such a great product without paying someone,
     please feel free to send a few bucks ($25) to support further
     development.

  6) This file was formated with tabs set to 2.

	Please forward any comments or suggestions to Bill Florac at:
	 	email: flash@etcconnect.com
		www: http://sumac.etcconnect.com/~fitco/
		mail: FITCO
					209 Jenna Dr
					Verona, WI  53593

	Revision History
		2/28/96
    	1.0 released
}


interface

uses
	SysUtils, WinProcs, WinTypes;

type
  	TDriveStyle = (tUnknown, tNoDrive, t3Floppy, t5Floppy, tFixed, tRFixed,
		tNetwork, tCDROM, tTape);

		PDeviceParams = ^TDeviceParams;
		TDeviceParams = record
			bSpecFunc: 			byte;			{Special functions}
			bDevType: 			byte;			{Device type}
			wDevAttr: 			word;			{Device attributes}
			wCylinders: 		word;			{Number of cylinders}
			bMediaType: 		byte;			{Media type}
{                          Beginning of BIOS parameter block (BPB)}
			wBytesPerSec: 	word;			{Bytes per sector}
			bSecPerClust: 	byte;			{Sectors per cluster}
			wResSectors: 		word;			{Number of reserved sectors}
			bFATs: 					byte;			{Number of FATs}
			wRootDirEnts: 	word;			{Number of root-directory entries}
			wSectors: 			word;			{Total number of sectors}
			bMedia: 				byte;			{Media descriptor}
			wFATsecs: 			word;			{Number of sectors per FAT}
			wSecPerTrack: 	word;			{Number of sectors per track}
			wHeads: 				word;			{Number of heads}
			dwHiddenSecs: 	longInt;	{Number of hidden sectors}
			dwHugeSectors: 	longInt;	{Number of sectors if wSectors == 0}
			reserved: 			array[0..10] of char;
{                          End of BIOS parameter block (BPB)}
		end;

	{parameter block for getting serial number}
	PSerialNumberParams = ^TSerialNumberParams;
	TSerialNumberParams = record
		wInfoLevel: 				word;
		dwDiskSerialNumber: longint;
		caLabel: 						array[0..10] of char;
		baFileSystem: 			array[0..7] of char;
	  end;


	{parameter block to get extened error codes}
	PExtErrorParams = ^TExtErrorParams;
  TExtErrorParams = record
    eCode: 		word;
		eClass: 	word;
		eAction: 	word;
		eLocus: 	word;
		eVolume: 	String;
		end;

	{structure for FCB}
  TEFCB = record
    Flag: 			byte;
    Reserved: 	array [0..4] of char;
    Attribute: 	byte;
    Drive: 			byte;
    Name: 			array [0..7] of char;
    Extension: 	array [0..2] of char;
    Misc: 			array  [0..24] of char;
	end;

DriveLabel = string[11];

  {my exception class}
	EDriveException = Class(Exception);

const
	{$I strings}

{standard calls}
function DriveReady(wDrive: word): boolean;
{Tests to see if a drive is ready.  (floppy there and door closed)}

function GetDriveLabel(wDrive: word): string;
function SetDriveLabel(wDrive: word; s: string): boolean;
{Gets and sets drive label}

function GetDriveSerialNumber(wDrive: word): LongInt;
function SetDriveSerialNumber(wDrive: word; SerialNumber: LongInt): boolean;
{Gets and sets drive serial number}

function GetDefaultDrive: word;
{Returns current default drive}

function GetDriveStyle(wDrive: word): TDriveStyle;
{Returns the drive style (hard, 3-1/2, 5-1/4...)}

procedure GetExtendedErrorInfo(ep: PExtErrorParams);
{Gets the parameters for the last DOS error.  Useful after a DriveReady failure.}

{other calls}
function IsCDROMDrive(wDrive: word): boolean;
function WriteDriveSNParam(wDrive: word; psnp: PSerialNumberParams): boolean;
function ReadDriveSNParam(wDrive: word; psnp: PSerialNumberParams): boolean;
function GetDeviceParameters(wDrive: word; var dp: TDeviceParams): boolean;


implementation

{determins if the drive is ready w/o critical errors enabled}
function DriveReady(wDrive: word): boolean;
var
  OldErrorMode: Word;
begin
	{turn off errors}
  OldErrorMode := SetErrorMode(SEM_FAILCRITICALERRORS);
	try
		if DiskSize(wDrive) = -1
		then result := false
		else result := true;
  finally
		{turn on errors}
		SetErrorMode(OldErrorMode);
	end;

end;

{get drive parameters w/o drive access}
function GetDeviceParameters(wDrive: word; var dp: TDeviceParams): boolean;
begin
	result := TRUE;      {Assume success}
   asm
			push ds
      mov  bx, wDrive
      mov  ch, 08h      {Device category--must be 08h}
      mov  cl, 60h      {MS-DOS IOCTL Get Device Parameters}
      lds  dx, dp
      mov  ax, 440Dh
      int  21h
      jnc  @gdp_done     {CF SET if error}
      mov  result, FALSE
   @gdp_done:
      pop  ds
	end;
end;

{gets last error message from DOS}
procedure GetExtendedErrorInfo(ep: PExtErrorParams);
var
	tCode: word;
	tClass: byte;
	tAction: byte;
	tLocus: byte;
begin
  asm
		push ds
		push bp
  	mov  bx, 0
  	mov  ah, 59h
    int  21h
		mov	 tCode, ax
		mov	 tClass, bh
    mov	 tAction, bl
		mov	 tLocus, ch
		pop  bp
		pop  ds
	end;
  ep^.eCode := tCode;
	ep^.eClass := tClass;
	ep^.eAction := tAction;
	ep^.eLocus := tLocus;
	ep^.eVolume := '?'; {don't support this for now}
end;

{get volume serial number for a drive:  0=default, 1=A...}
{returns -1 if unable to read}
function GetDriveSerialNumber(wDrive: word): LongInt;
var
	snp: TSerialNumberParams;
begin
	snp.dwDiskSerialNumber := 0;
	if ReadDriveSNParam(wDrive, @snp)
	then Result := snp.dwDiskSerialNumber
	else Result := -1;
end;


{set volume serial number for a drive:  0=default, 1=A... }
{returns true if it was sucessful}
function SetDriveSerialNumber(wDrive: word; SerialNumber: LongInt): boolean;
var
	snp: TSerialNumberParams;
begin
	result := false;
	{get current parameters}
	if ReadDriveSNParam(wDrive, @snp) then begin
		{change serial number}
		snp.dwDiskSerialNumber := SerialNumber;
		{and write back out}
		if WriteDriveSNParam(wDrive, @snp) then result := true;
	end;
end;

{Write Drive parameters: 0=default, 1=A...}
{Note: wDrive and psnp are treate as var with assembler directive}
{This interupt does NOT generate a critical error!}
function WriteDriveSNParam(wDrive: word; psnp: PSerialNumberParams): boolean; assembler;
asm
	push ds           {ds might get changed so save it}
	mov  bx, wDrive
  mov  al, 01h
	mov  ah, 69h
  lds  dx, psnp
  int  21h
  jnc  @no_error    {CF SET if error}
	xor	 ax,ax				{set false}
  jmp	 @exit
@no_error:
  mov	ax, 1					{set true}
@exit:
	pop	ds						{restore ds}
end;

{Read Drive parameters: 0=default, 1=A...}
{Note: wDrive and psnp are treate as var with assembler directive}
{This interupt does NOT generate a critical error!}
function ReadDriveSNParam(wDrive: word; psnp: PSerialNumberParams): boolean; assembler;
asm
	push ds
	mov  bx, wDrive
  mov  al, 00h
	mov  ah, 69h
  lds  dx, psnp
  int  21h
  jnc  @no_error    	{CF SET if error}
	xor	 ax,ax		{set false}
  jmp	 @exit
@no_error:
  mov	ax, 1			{set true}
@exit:
	pop	ds
end;

{sets the label of the drive specified: wDrive: 0=default 1=A...}
{returns true if it was sucessful}
function SetDriveLabel(wDrive: word; s: string): boolean;
const
		EFCB: TEFCB = (
	    Flag: $FF;                          	{ Extended FCB Flag }
      Reserved: (#0,#0,#0,#0,#0);           { Reserved}
      Attribute: $08;                       { Volume Label Attribute}
      Drive: 2;                             { Drive Identifier}
      Name: '????????';  										{ File Name}
      Extension: '???';                     { File Extension}
      Misc: (#0, #0, #0, #0, #0,            { Misc. Info filled by DOS}
        ' ',' ',' ',' ',' ',' ',' ',' ',  	{ Misc. Info filled by DOS}
        ' ',' ',' ',                      	{ Misc. Info filled by DOS}
        #0, #0, #0, #0, #0, #0, #0, #0, #0  { Misc. Info filled by DOS}
				)
      );
var
	Ps: pchar;
	err: integer;
	x: integer;

begin
	{abort if drive not ready}
	if not DriveReady(wDrive) then begin
		result := false;
		exit;
	end;
	{assume ok}
	result := true;

	{default things that change in constant varaiable}
	EFCB.Name := '????????';
	EFCB.Extension := '???';
	EFCB.Drive := wDrive;

	{See if it exist using a FCB}
	asm
		{Check to see;  if the volume label exists}
		{point DTA to ourself}
		mov	 dx,offset EFCB
		mov	 ah,1Ah
		int  21h
		{point to default FCB}
   	mov  dx, offset EFCB
    mov  ah, 11h
    int  21h
		{Exit if label is not present}
    cmp  al, 0
    jne  @exit
		{Else delete the volume label}
    mov  dx, offset EFCB
    mov  ah, 013h
    int  21h
    or	 al,al
		jz	@exit
		mov	result, 0
  @exit:
	end;

	if not result then exit;

	{if string is empty, then just erase}
	if length(s) = 0 then exit;
	{format string}
	for x := length(s) + 1 to 11  do s[x] := ' ';
	s[0] := char(11);
	{add drive letter!}
  if wdrive = 0
	then s := '\' + s + #0
	else s := chr(64+wdrive) + ':\' + s + #0;
	ps := @s[1];

	{on now make new one it!}
	asm
		push ds
    lds  dx, ps
    mov  cx, faVolumeID
    mov  ah,3Ch
    int  21h
    {CF set if error}
    jnc  @noerror
    mov  result, FALSE
		jmp	 @exit
  @noerror:
		{close file ax = handle}
		mov  bx,ax
		mov	 ah,3Eh
		int	 21h
  @exit:
  	pop  ds
	end
end;

{Get label from drive.  0=default, 1=A...}
{return string of 11 character or "NO NAME" if not found}
function GetDriveLabel(wDrive: word): string;
const
	pattern: string[6] = 'c:\*.*';
var
	sr: TsearchRec;
  OldErrorMode: Word;
  DotPos: Byte;
begin
	{get default drive}
	if wDrive = 0
	then wDrive := GetDefaultDrive
	else dec(wDrive);

	{switch out drive letter}
	pattern[1] := char(65 + wDrive);

	{stop errors and try}
  OldErrorMode := SetErrorMode(SEM_FAILCRITICALERRORS);
	try
		if FindFirst(Pattern, faVolumeID, sr) = 0 then begin
      Result := sr.Name;
      DotPos := Pos('.', Result);
      if DotPos <> 0 then Delete(Result, DotPos, 1);
	  end
		else result := 'NO NAME'
  finally
		{restore errorsa}
		SetErrorMode(OldErrorMode);
  end;
end;

function GetDefaultDrive: word; assembler;
asm
	mov	ah, 19h 			{convert default to real}
	int	21h
	xor	ah, ah				{clear hi byte}
end;

{Determine id drive is a CDROM, 0=default, 1=A ...}
function IsCDROMDrive(wDrive: word): boolean; assembler;
var
	wTempDrive: word;
asm
	mov	ax, wDrive
	or  ax, ax
	jnz	@not_default
	mov	ah, 19h 			{convert default to drive}
	int	21h
	xor	ah, ah
	mov wTempDrive, ax
	jmp	@test_it
@not_default: 			{zero base it}
	dec	ax
	mov wTempDrive, ax
@test_it:
	mov ax, 1500h     {first test for presence of MSCDEX}
  xor bx, bx
  int 2fh
  mov ax, bx        {MSCDEX is not there if bx is zero}
	or  ax, ax        {so return FALSE}
	jz  @no_mscdex
	mov ax, 150bh     {MSCDEX driver check API}
	mov cx, wTempDrive    {...cx is drive index}
	int 2fh
	or 	ax, ax
@no_mscdex:
end;

{returns drive type}
{read BOIS not drive so floppy does not have to be in drive}
{I don't have all types of drive so not all could be tested}
function GetDriveStyle(wDrive: word): TDriveStyle;
var
	x: word;
	wTempDrive: word;
	dp: TDeviceParams;
begin
	{convert default to drive}
	if wDrive = 0
	then wTempDrive := GetDefaultDrive
	else wTempDrive := wDrive - 1;
	x := GetDriveType(wTempDrive);

	{get types}
	case x of
	drive_Removable: begin
		dp.bSpecFunc := 0; {need to clear this}
		if GetDeviceParameters(wDrive,dp) then begin
			case dp.bDevType of
			0,1: result := t5floppy;		{320K/360K/1.2M}
			2,7,9: result := t3floppy;	{720K/1.44M/2.88M}
			5: result := tRFixed;     	{yes a removable fixed drive!}
			6: result := tTape;         {tape}
      else result := tUnknown;
			end;
		end
		else result := tUnknown;
	end;
  drive_Fixed:
		if IsCDROMDrive(wDrive)
		then result := tCDROM
		else result := tFixed;
 	drive_Remote:
		if IsCDROMDrive(wDrive) {I think this is possible on a network!}
		then result := tCDROM
		else result := tNetWork;
	else result := tUnknown;
	end;
end;

end. {of unit}
